/*
 * Copyright 2017-present Open Networking Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.onosproject.fpcagent.protocols;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.json.JSONObject;
import org.onlab.packet.Ip4Address;
import org.onlab.packet.Ip4Prefix;
import org.onosproject.fpcagent.util.CacheManager;
import org.onosproject.fpcagent.workers.ZmqSbPublisherManager;
import org.onosproject.fpcagent.workers.ZmqSbSubscriberManager;
import org.onosproject.yang.gen.v1.fpc.rev20150105.fpc.ZmqDpnControlProtocol;
import org.onosproject.yang.gen.v1.ietfdmmfpcagent.rev20160803.ietfdmmfpcagent.tenants.tenant.fpctopology.DefaultDpns;
import org.onosproject.yang.gen.v1.ietfdmmfpcbase.rev20160803.ietfdmmfpcbase.FpcDpnId;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Optional;

import static org.onosproject.fpcagent.util.Converter.*;
import static org.onosproject.fpcagent.util.FpcUtil.getTopicFromNode;

/**
 * DPDK DPN API over ZeroMQ for NGIC.
 */
public final class DpnNgicCommunicator extends ZmqDpnControlProtocol implements DpnCommunicationService {
    protected static final Logger log = LoggerFactory.getLogger(DpnNgicCommunicator.class);

    private static DpnNgicCommunicator _instance;

    private DpnNgicCommunicator() {
    }

    public static DpnNgicCommunicator createInstance() {
        if (_instance == null) {
            _instance = new DpnNgicCommunicator();
        }
        return _instance;
    }

    public static DpnNgicCommunicator getInstance() {
        return _instance;
    }

    /**
     * Broadcasts the GOODBYE message to all the DPNs.
     *
     * @param nodeId node identifier
     * @param networkId network identifier
     */
    public static void sendGoodbyeDpns(String nodeId, String networkId) {
        ByteBuffer bb = ByteBuffer.allocate(10 + nodeId.length() + networkId.length());
        bb.put(ReservedTopics.BROADCAST_DPNS.getType())
                .put(S11MsgType.CONTROLLER_STATUS_INDICATION.getType())
                .put(ZmqSbSubscriberManager.getInstance().getControllerTopic())
                .put(ControllerStatusIndication.GOODBYE.getType())
                .put(toUint32(ZmqSbSubscriberManager.getInstance().getControllerSourceId()))
                .put(toUint8((short) nodeId.length()))
                .put(nodeId.getBytes())
                .put(toUint8((short) networkId.length()))
                .put(networkId.getBytes());

        log.info("sendGoodbyeDpns: {}", bb.array());
        ZmqSbPublisherManager.getInstance().send(bb);
    }

    /**
     * Broadcasts the HELLO message to all the DPNs.
     *
     * @param nodeId node identifier
     * @param networkId network identifier
     */
    public static void sendHelloDpns(String nodeId, String networkId) {
        ByteBuffer bb = ByteBuffer.allocate(10 + nodeId.length() + networkId.length());
        bb.put(ReservedTopics.BROADCAST_DPNS.getType())
                .put(S11MsgType.CONTROLLER_STATUS_INDICATION.getType())
                .put(ZmqSbSubscriberManager.getInstance().getControllerTopic())
                .put(ControllerStatusIndication.HELLO.getType())
                .put(toUint32(ZmqSbSubscriberManager.getInstance().getControllerSourceId()))
                .put(toUint8((short) nodeId.length()))
                .put(nodeId.getBytes())
                .put(toUint8((short) networkId.length()))
                .put(networkId.getBytes());

        log.info("sendHelloDpns: {}", bb.array());
        ZmqSbPublisherManager.getInstance().send(bb);
    }

    /**
     * Broadcasts a conflict on node and network identifiers.
     *
     * @param nodeId node identifier
     * @param networkId network identifier
     */
    public static void sendAssignConflict(String nodeId, String networkId) {
        ByteBuffer bb = ByteBuffer.allocate(9 + nodeId.length() + networkId.length());
        bb.put(ReservedTopics.BROADCAST_ALL.getType())
                .put(S11MsgType.ASSIGN_CONFLICT.getType())
                .put(ZmqSbSubscriberManager.getInstance().getControllerTopic())
                .put(toUint32(ZmqSbSubscriberManager.getInstance().getControllerSourceId()))
                .put(toUint8((short) nodeId.length()))
                .put(nodeId.getBytes())
                .put(toUint8((short) networkId.length()))
                .put(networkId.getBytes());

        log.info("sendAssignConflict: {}", bb.array());
        ZmqSbPublisherManager.getInstance().send(bb);
    }

    /**
     * Broadcasts a topic to be assigned for the specific node and network.
     *
     * @param nodeId node identifier
     * @param networkId network identifier
     * @param topic assigned topic
     */
    public static void sendAssignTopic(String nodeId, String networkId, byte topic) {
        ByteBuffer bb = ByteBuffer.allocate(9 + nodeId.length() + networkId.length());
        bb.put(ReservedTopics.BROADCAST_ALL.getType())
                .put(S11MsgType.ASSIGN_TOPIC.getType())
                .put(topic)
                .put(toUint32(ZmqSbSubscriberManager.getInstance().getControllerSourceId()))
                .put(toUint8((short) nodeId.length()))
                .put(nodeId.getBytes())
                .put(toUint8((short) networkId.length()))
                .put(networkId.getBytes());

        log.info("sendAssignTopic: {}", bb.array());
        ZmqSbPublisherManager.getInstance().send(bb);
    }

    /**
     * Sends acknowledgment for the DPN status.
     *
     * @param nodeId node identifier
     * @param networkId network identifier
     * @param topic node's topic
     */
    public static void sendStatusAck(String nodeId, String networkId, byte topic) {
        ByteBuffer bb = ByteBuffer.allocate(9 + nodeId.length() + networkId.length())
                .put(topic)
                .put(S11MsgType.DPN_STATUS_ACK.getType())
                .put(ZmqSbSubscriberManager.getInstance().getControllerTopic())
                .put(toUint32(ZmqSbSubscriberManager.getInstance().getControllerSourceId()))
                .put(toUint8((short) nodeId.length()))
                .put(nodeId.getBytes())
                .put(toUint8((short) networkId.length()))
                .put(networkId.getBytes());

        log.info("sendStatusAck: {}", bb.array());
        ZmqSbPublisherManager.getInstance().send(bb);
    }

    /**
     * Parses the JSON returned from the Control Plane and sends the ACK to the DPN.
     *
     * @param body - The JSON body returned by the Control Plane in the DDN ACK
     */
    public static void sendDdnAck(JSONObject body) {
        try {
            ByteBuffer bb = ByteBuffer.allocate(14);
            byte topic = -1;
            for (CacheManager cacheManager : CacheManager.cacheInfo.values()) {
                final Optional<DefaultDpns> optionalDpn = cacheManager.dpnsCache.get(
                        FpcDpnId.fromString((String) body.get("dpn-id"))
                );

                if (optionalDpn.isPresent()) {
                    DefaultDpns dpn = optionalDpn.get();
                    topic = getTopicFromNode(dpn.nodeId() + "/" + dpn.networkId());
                    break;
                }
            }

            if (topic == -1) {
                throw new RuntimeException("DPN specified in DDN not found.");
            }

            bb.put(topic).put(S11MsgType.DDN_ACK.getType());

            if (body.has("dl-buffering-duration")) {
                bb.put(toUint8((short) body.getInt("dl-buffering-duration")));
            }
            if (body.has("dl-buffering-suggested-count")) {
                bb.put(toUint16((int) body.getInt("dl-buffering-suggested-count")));
            }
            bb.put(ZmqSbSubscriberManager.getInstance().getControllerTopic())
                    .put(toUint32(Long.parseLong(body.getString("client-id"))))
                    .put(toUint32(body.getLong("op-id")));

            ZmqSbPublisherManager.getInstance().send(bb);
        } catch (Exception e) {
            log.error("{}", ExceptionUtils.getFullStackTrace(e));
        }

    }

    @Override
    public void createSession(
            byte topicId,
            BigInteger imsi,
            Short defaultEbi,
            Ip4Address ueIpv4,
            Long s1USgwTeid,
            Ip4Address s1USgwIpv4,
            Long sessionId,
            Long clientId,
            BigInteger opId
    ) {
        /*
            NGIC Create Session expected buffer:
            uint8_t topic_id;
            uint8_t type;
            struct create_session_t {
                uint64_t imsi;
                uint8_t  default_ebi;
                uint32_t ue_ipv4;
                uint32_t s1u_sgw_teid;
                uint32_t s1u_sgw_ipv4;
                uint64_t session_id;
                uint8_t  controller_topic;
                uint32_t client_id;
                uint32_t op_id;
            } create_session_msg;
         */
        ByteBuffer bb = ByteBuffer.allocate(41)
                .put(topicId)
                .put(S11MsgType.CREATE_SESSION.getType())
                .put(toUint64(imsi))
                .put(toUint8(defaultEbi))
                .put(toUint32(ueIpv4.toInt()))
                .put(toUint32(s1USgwTeid))
                .put(toUint32(s1USgwIpv4.toInt()))
                .put(toUint64(BigInteger.valueOf(sessionId)))
                .put(ZmqSbSubscriberManager.getInstance().getControllerTopic())
                .put(toUint32(clientId))
                .put(toUint32(opId.longValue()));

        log.info("createSession(topic_id={}, imsi={}, default_ebi={}, " +
                        "ue_ipv4={}, s1u_sgw_teid={}, s1u_sgw_ipv4={}, " +
                        "session_id={}, client_id={}, op_id={})",
                topicId, imsi, defaultEbi, ueIpv4, s1USgwTeid,
                s1USgwIpv4, sessionId, clientId, opId);

        ZmqSbPublisherManager.getInstance().send(bb);
    }

    @Override
    public void modifyBearer(
            byte topicId,
            Ip4Address s1USgwIpv4,
            Long s1UEnodebTeid,
            Ip4Address s1UEnodebIpv4,
            Long sessionId,
            Long clientId,
            BigInteger opId
    ) {
        /*
            NGIC Modify Session expected buffer:
            uint8_t topic_id;
            uint8_t type;
            struct modify_bearer_t {
                uint32_t s1u_sgw_ipv4;
                uint32_t s1u_enodeb_teid;
                uint32_t s1u_enodeb_ipv4;
                uint64_t session_id;
                uint8_t  controller_topic;
                uint32_t client_id;
                uint32_t op_id;
            } modify_bearer_msg;
         */
        ByteBuffer bb = ByteBuffer.allocate(32)
                .put(topicId)
                .put(S11MsgType.UPDATE_MODIFY_BEARER.getType())
                .put(toUint32(s1USgwIpv4.toInt()))
                .put(toUint32(s1UEnodebTeid))
                .put(toUint32(s1UEnodebIpv4.toInt()))
                .put(toUint64(BigInteger.valueOf(sessionId)))
                .put(ZmqSbSubscriberManager.getInstance().getControllerTopic())
                .put(toUint32(clientId))
                .put(toUint32(opId.longValue()));

        log.info("modifyBearer(topic_id={}, s1u_sgw_ipv4={}, " +
                        "s1u_enodeb_teid={}, s1u_enodeb_ipv4={}, " +
                        "session_id={}, client_id={}, op_id={})",
                topicId, s1USgwIpv4, s1UEnodebTeid, s1UEnodebIpv4,
                sessionId, clientId, opId);

        ZmqSbPublisherManager.getInstance().send(bb);
    }

    @Override
    public void deleteSession(
            byte topicId,
            Long sessionId,
            Long clientId,
            BigInteger opId
    ) {
        /*
            NGIC Delete Session expected buffer:
            uint8_t topic_id;
            uint8_t type;
            struct delete_session_t {
                uint64_t session_id;
                uint8_t  controller_topic;
                uint32_t client_id;
                uint32_t op_id;
            } delete_session_msg;
         */
        ByteBuffer bb = ByteBuffer.allocate(19)
                .put(topicId)
                .put(S11MsgType.DELETE_SESSION.getType())
                .put(toUint64(BigInteger.valueOf(sessionId)))
                .put(ZmqSbSubscriberManager.getInstance().getControllerTopic())
                .put(toUint32(clientId))
                .put(toUint32(opId.longValue()));

        log.info("deleteSession(topic_id={}, session_id={}, client_id={}, " +
                        "op_id={})",
                topicId, sessionId, clientId, opId);

        ZmqSbPublisherManager.getInstance().send(bb);
    }

    @Override
    public void sendAdcRules(
            Short topic,
            String domainName,
            String ip,
            Short drop,
            Long ratingGroup,
            Long serviceId, String sponsorId
    ) {
        // TODO take a look for this function. Not tested.
        Ip4Prefix ip4Prefix = null;
        if (ip != null) {
            ip4Prefix = Ip4Prefix.valueOf(ip);
        }
        Short selectorType = (short) (domainName != null ? 0 : ip4Prefix != null ?
                2 : ip4Prefix.address() != null ? 1 : 255);
        if (selectorType == 255) {
            log.warn("Domain/IP not found, failed to send rules");
            return;
        }
        ByteBuffer bb = ByteBuffer.allocate(200);
        bb.put(toUint8(topic))
                .put(S11MsgType.ADC_RULE.getType())
                .put(toUint8(selectorType));
        if (selectorType == 0) {
            bb.put(toUint8((short) domainName.length()))
                    .put(domainName.getBytes());
        }
        if ((selectorType == 1) || (selectorType == 2)) {
            int ipAddressLong = ip4Prefix.address().toInt();
            bb.put(toUint32(ipAddressLong));
        }
        if (selectorType == 2) {
            bb.put(toUint16(ip4Prefix.prefixLength()));
        }
        if (drop != null) {
            bb.put(toUint8(drop));
        }
        if (ratingGroup != null) {
            bb.put(toUint32(ratingGroup));
        }
        if (serviceId != null) {
            bb.put(toUint32(serviceId));
        }
        if (sponsorId != null && (short) sponsorId.length() > 0) {
            bb.put(toUint8((short) sponsorId.length()))
                    .put(sponsorId.getBytes());
        }
        bb.put(ZmqSbSubscriberManager.getInstance().getControllerTopic());

        log.info("sendAdcRules: {}", bb.array());
        ZmqSbPublisherManager.getInstance().send(bb);
    }

    public enum S11MsgType {
        CREATE_SESSION(1) {
            @Override
            public String toString() {
                return "CREATE_SESSION";
            }
        },
        UPDATE_MODIFY_BEARER(2) {
            @Override
            public String toString() {
                return "UPDATE_MODIFY_BEARER";
            }
        },
        DELETE_SESSION(3) {
            @Override
            public String toString() {
                return "DELETE_SESSION";
            }
        },
        DPN_RESPONSE(4) {
            @Override
            public String toString() {
                return "DPN_RESPONSE";
            }
        },
        DDN(5) {
            @Override
            public String toString() {
                return "DDN";
            }
        },
        DDN_ACK(5) {
            @Override
            public String toString() {
                return "DDN_ACK";
            }
        },
        RESERVED_1(7) {
            @Override
            public String toString() {
                return "RESERVED_1";
            }
        },
        RESERVED_2(8) {
            @Override
            public String toString() {
                return "RESERVED_2";
            }
        },
        RESERVED_3(9) {
            @Override
            public String toString() {
                return "RESERVED_3";
            }
        },
        ASSIGN_TOPIC(10) {
            @Override
            public String toString() {
                return "ASSIGN_TOPIC";
            }
        },
        ASSIGN_CONFLICT(11) {
            @Override
            public String toString() {
                return "ASSIGN_CONFLICT";
            }
        },
        DPN_STATUS_INDICATION(12) {
            @Override
            public String toString() {
                return "DPN_STATUS_INDICATION";
            }
        },
        DPN_STATUS_ACK(13) {
            @Override
            public String toString() {
                return "DPN_STATUS_ACK";
            }
        },
        CONTROLLER_STATUS_INDICATION(14) {
            @Override
            public String toString() {
                return "CONTROLLER_STATUS_INDICATION";
            }
        },
        GENERATE_CDR(15) {
            @Override
            public String toString() {
                return "GENERATE_CDR";
            }
        },
        GENERATE_CDR_ACK(16) {
            @Override
            public String toString() {
                return "GENERATE_CDR_ACK";
            }
        },
        ADC_RULE(17) {
            @Override
            public String toString() {
                return "ADC_RULE";
            }
        },
        PCC_RULE(18) {
            @Override
            public String toString() {
                return "PCC_RULE";
            }
        },
        METER_RULE(19) {
            @Override
            public String toString() {
                return "METER_RULE";
            }
        },
        SDF_RULE(20) {
            @Override
            public String toString() {
                return "SDF_RULE";
            }
        };

        private byte type;

        S11MsgType(int type) {
            this.type = (byte) type;
        }

        public static S11MsgType getEnum(byte name) {
            Optional<S11MsgType> any = Arrays.stream(S11MsgType.values())
                    .filter(typeStr -> typeStr.type == name)
                    .findAny();
            if (any.isPresent()) {
                return any.get();
            }
            throw new IllegalArgumentException("No enum defined for string: " + name);
        }

        public byte getType() {
            return type;
        }
    }

    public enum DpnStatusIndication {
        HELLO(1) {
            @Override
            public String toString() {
                return "HELLO";
            }
        },
        GOODBYE(2) {
            @Override
            public String toString() {
                return "GOODBYE";
            }
        },
        OVERLOAD_START(3) {
            @Override
            public String toString() {
                return "OVERLOAD_START";
            }
        },
        OVERLOAD_STOP(4) {
            @Override
            public String toString() {
                return "OVERLOAD_STOP";
            }
        },
        MATERIAL_CHANGE(5) {
            @Override
            public String toString() {
                return "MATERIAL_CHANGE";
            }
        },
        RESTART(6) {
            @Override
            public String toString() {
                return "RESTART";
            }
        },
        OUT_OF_SERVICE(7) {
            @Override
            public String toString() {
                return "OUT_OF_SERVICE";
            }
        };

        private byte type;

        DpnStatusIndication(int type) {
            this.type = (byte) type;
        }

        public static DpnStatusIndication getEnum(byte name) {
            Optional<DpnStatusIndication> any = Arrays.stream(DpnStatusIndication.values())
                    .filter(typeStr -> typeStr.type == name)
                    .findAny();
            if (any.isPresent()) {
                return any.get();
            }
            throw new IllegalArgumentException("No enum defined for string: " + name);
        }

        public byte getType() {
            return type;
        }
    }

    public enum ControllerStatusIndication {
        HELLO(1) {
            @Override
            public String toString() {
                return "HELLO";
            }
        },
        GOODBYE(2) {
            @Override
            public String toString() {
                return "GOODBYE";
            }
        };

        private byte type;

        ControllerStatusIndication(int type) {
            this.type = (byte) type;
        }

        public byte getType() {
            return type;
        }
    }

    public enum ReservedTopics {
        BROADCAST_ALL(1) {
            @Override
            public String toString() {
                return "BROADCAST_ALL";
            }
        },
        BROADCAST_CONTROLLERS(2) {
            @Override
            public String toString() {
                return "BROADCAST_CONTROLLERS";
            }
        },
        BROADCAST_DPNS(3) {
            @Override
            public String toString() {
                return "BROADCAST_DPNS";
            }
        };

        private byte type;

        ReservedTopics(int type) {
            this.type = (byte) type;
        }

        public byte getType() {
            return type;
        }
    }

}