/*
 * Copyright 2017-present Open Networking Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.onosproject.fpcagent.workers;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.zeromq.ZContext;
import org.zeromq.ZMQ;

import java.nio.ByteBuffer;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;

public class ZmqSbPublisherManager implements AutoCloseable {
    private static final Logger log = LoggerFactory.getLogger(ZmqSbPublisherManager.class);
    private static ZmqSbPublisherManager _instance = null;
    private final ZContext ctx;
    private final String address;
    private final BlockingQueue<ByteBuffer> blockingQueue;
    private final int poolSize;
    private boolean run;

    protected ZmqSbPublisherManager(String address, int poolSize) {
        this.ctx = new ZContext();
        this.address = address;
        this.run = true;
        this.blockingQueue = new LinkedBlockingQueue<>();
        this.poolSize = poolSize;
    }

    public static ZmqSbPublisherManager createInstance(String address, int poolSize) {
        if (_instance == null) {
            _instance = new ZmqSbPublisherManager(address, poolSize);
        }
        return _instance;
    }

    public static ZmqSbPublisherManager getInstance() {
        return _instance;
    }

    public void send(ByteBuffer buf) {
        try {
            blockingQueue.put(buf);
        } catch (InterruptedException e) {
            log.error(ExceptionUtils.getFullStackTrace(e));
        }
    }

    public void open() {
        Executors.newFixedThreadPool(this.poolSize).submit(() -> {
            ZMQ.Socket socket = ctx.createSocket(ZMQ.PUB);
            socket.connect(address);
            log.debug("Publisher at {}", address);
            while ((!Thread.currentThread().isInterrupted()) && run) {
                try {
                    byte[] array = blockingQueue.take().array();
                    socket.send(array);
                } catch (InterruptedException e) {
                    log.error(ExceptionUtils.getFullStackTrace(e));
                }
            }
        });
    }

    @Override
    public void close() {
        run = false;
    }
}
